/*
 *  $Id: streamer.c,v 1.19 2009/06/19 06:32:45 schmirl Exp $
 */
 
#include <vdr/ringbuffer.h>
#include <vdr/device.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#include "server/streamer.h"
#include "server/livestreamer.h"
#include "server/suspend.h"
#include "server/setup.h"
#include "tools/socket.h"
#include "tools/select.h"
#include "common.h"
#include <vdr/remux.h>

static void wait_microseconds(uint64_t  microseconds)                                                                                                                                                              
{                                                                                                                                                                                                             
#ifdef _WIN32                                                                                                                                                                                                 
  // Best we can (easily) do is to wait for the nearest (rounded down!)                                                                                                                                       
  // number of milliseconds - hopefully this will do                                                                                                                                                          
  Sleep(microseconds / 1000);                                                                                                                                                                                 
#else // _WIN32                                                                                                                                                                                               
  struct  timespec   time = {0};                                                                                                                                                                              
  struct  timespec   remaining;                                                                                                                                                                               
  uint32_t nanoseconds = microseconds * 1000;                                                                                                                                                                 
  int     err = 0;                                                                                                                                                                                            
                                                                                                                                                                                                              
  time.tv_sec = 0;                                                                                                                                                                                            
  time.tv_nsec = nanoseconds;                                                                                                                                                                                 
                                                                                                                                                                                                              
  errno = 0;                                                                                                                                                                                                  
  err = nanosleep(&time,&remaining);                                                                                                                                                                          
  while (err == -1 && errno == EINTR)  // cope with being woken too early                                                                                                                                     
  {                                                                                                                                                                                                           
    errno = 0;                                                                                                                                                                                                
    time = remaining;                                                                                                                                                                                         
    err = nanosleep(&time,&remaining);                                                                                                                                                                        
  }                                                                                                                                                                                                           
#endif // _WIN32                                                                                                                                                                                              
  return;                                                                                                                                                                                                     
}                                                                                                                                                                                                             

// --- cStreamdevBuffer -------------------------------------------------------

cStreamdevBuffer::cStreamdevBuffer(int Size, int Margin, bool Statistics, const char *Description):
		cRingBufferLinear(Size, Margin, Statistics, Description)
{
    letter = 'a'+rand() % 10;
}

// --- cStreamdevWriter -------------------------------------------------------

cStreamdevWriter::cStreamdevWriter(cTBSocket *Socket, 
                                   cStreamdevStreamer *Streamer):
		cThread("streamdev-writer"),
		m_Streamer(Streamer),
		m_Socket(Socket)
{
	writing = 0;
	sel = new cTBSelect;
	sel->Clear();
	sel->Add(*m_Socket, true);
    t1 = 0;
    t2 = 0;
}

cStreamdevWriter::~cStreamdevWriter()
{
	Dprintf("destructing writer\n");
	//close(fdw);
	if (Running())
		Cancel(3);
}

void cStreamdevWriter::Action(void)
{
//	cTBSelect sel;
	Dprintf("Writer start\n");
	int max = 0;
	uchar *block = NULL;
	int count, offset = 0;
	char fn[50];

/*	sprintf(fn, "%c.ts", m_Streamer->m_SendBuffer->letter);
	fdw = open(fn, O_WRONLY | O_CREAT | O_TRUNC);
	Dprintf("fd: %d, %s, %m", fdw, fn);*/

//	sel.Clear();
//	sel.Add(*m_Socket, true);
	
	uint64_t prev_pcr = 0;
	uint64_t prev_time = 0;
	uint64_t prev_log = 0;
	uint64_t curtime = 0;
	uint64_t first_pcr = 0;
	uint64_t first_time = 0;
	int refills = 0;
	int resets = 0;
	int num_overruns = 0;
	int need_flush = 0;
	int flushes = 0;
	int num_scr = 0;
	int num_descr = 0;
	while (Running()) {

		if (m_Streamer->m_SendBuffer->Available()<100000) {
			while (m_Streamer->m_SendBuffer->Available()<1000000) {
				usleep(40000);
			}
			Dprintf("Filled buffer!\n");
			refills++;
		}
		int avail = m_Streamer->m_SendBuffer->Available();
		if (avail > 3000000) {
			need_flush=1;
			flushes++;
		}
		if (avail < 1000000)
			need_flush=0;
		
		block = m_Streamer->Get(count);
		//if (m_Streamer->m_SendBuffer->Available() <100000) 
		//	Dprintf("%d\n", m_Streamer->m_SendBuffer->Available());
		offset = 0;

		if (block != NULL) {
			if (sel->Select(15000) == -1) {
				esyslog("ERROR: streamdev-server: couldn't send data: %m");
				break;
			}

			if (sel->CanWrite(*m_Socket)) {
				int written;
				int pkgsize = count;
				// SOCK_DGRAM indicates multicast
				if (m_Socket->Type() == SOCK_DGRAM) {
					// don't fragment multicast packets
					// max. payload on standard local ethernet is 1416 to 1456 bytes
					// and some STBs expect complete TS packets
					// so let's always limit to 7 * TS_SIZE = 1316
					if (pkgsize > 7 * TS_SIZE)
						pkgsize = 7 * TS_SIZE;
					else
						pkgsize -= pkgsize % TS_SIZE;
				}
				int blockpos = 0;
				while (blockpos < pkgsize) {
					int pid = TsPid(block+blockpos);
					if ((!need_flush) && (m_Streamer->pcrPid) && (m_Streamer->pcrPid == pid) && (TsGetAdaptationField(block+blockpos) & TS_ADAPT_PCR)) {
						/*if (pid == 3321) 
						Dprintf("1 pid: %d, first_pcr: %llu, start: %llu\n", 
								pid, first_pcr, first_time);*/

						struct timeval now;
						unsigned char *adapt = block + blockpos + 5;
						uint64_t pcr = ((uint64_t)adapt[1] << 25) | (adapt[2] << 17) | (adapt[3] << 9) | (adapt[4] << 1) | (adapt[5] >> 7);                                                                                                                                                                      
						// Plus the program clock reference extension                                                                                                                                                             
						pcr = ((pcr) * 300) + ((adapt[5] & 1) << 8) + adapt[6];
						gettimeofday(&now, NULL);
						curtime = (uint64_t)now.tv_sec*1000000 + (uint64_t)now.tv_usec;
						/*if (pid == 3321) 
						Dprintf("2 pid: %d, first_pcr: %llu, start: %llu\n", 
								pid, first_pcr, first_time);*/

						if (first_pcr == 0)
							first_pcr = pcr;
						if (first_time == 0)
							first_time = curtime;

						int64_t elapsed_time = curtime-first_time;
						int64_t pcr_diff = (pcr-first_pcr)/27;
						int64_t need_sleep = pcr_diff-elapsed_time;

						if (avail > 1400000) {
							need_sleep -= 1000;
						} else if (avail < 700000) {
							need_sleep += 1000;
						}
						
						prev_time = curtime;
						prev_pcr = pcr;
						if ((need_sleep>0) && (need_sleep<200000)) {
							wait_microseconds(need_sleep);
							prev_time += need_sleep;
							num_overruns=0;
						} else {
							num_overruns++;
						}

						if (num_overruns > 5) {
							first_time = curtime;
							first_pcr = pcr;
							num_overruns = 0;
							resets++;
						}
						
						//if (pid == 3321) 
						//Dprintf("pid: %d, pcr: %llu, start: %llu, now: %llu, elapsed_time: %lld, pcr_diff: %lld, sleep: %lld, avail %d\n", 
						//		pid, pcr/27, first_time, curtime, elapsed_time, pcr_diff, need_sleep, m_Streamer->m_SendBuffer->Available());
					}
					blockpos += TS_SIZE;
				}
				
				if (block[3] & TS_SCRAMBLING_CONTROL) {
					num_scr++;
				} else {
					num_descr++;
				}
				
//				if ((written = m_Socket->Write(block + offset, pkgsize)) == -1) {
//				int w = write(fdw, block + offset, pkgsize);
//				Dprintf("%d, %m\n", w);
				if ((written =     send (*m_Socket, block + offset, pkgsize, MSG_DONTWAIT | MSG_EOR)) == -1) {
					esyslog("ERROR: streamdev-server: couldn't send %d bytes: %m", pkgsize);
					break;
				}
				// statistics
				if (count > max)
					max = count;

				offset += written;
				count -= written;

				// less than one TS packet left:
				// delete what we've written so far and get next chunk
				//if (count < TS_SIZE) {
					m_Streamer->Del(offset);
					block = NULL;
				//}
				//				usleep(1);
				if (curtime-prev_log > 20000000) {
					cStreamdevLiveStreamer *stm = (cStreamdevLiveStreamer*)m_Streamer;
					Dprintf("Channel %s, scrambled: %d, descrambled: %d, buffer available: %d, refills: %d, resets: %d, flushes: %d\n",
						stm->GetChannel()->Name(), num_scr, num_descr, m_Streamer->m_SendBuffer->Available(), refills, resets, flushes);
					prev_log = curtime;
				}
			}
		}
	}
	Dprintf("Max. Transmit Blocksize was: %d\n", max);
}

void cStreamdevWriter::ActionCall(void)
{
	int max = 0;
	uchar *block = NULL;
	int count, offset = 0;
	int needlog = 0;
	timeval tim;
	
	if (writing) {
		Dprintf("writer already writing\n");
		return;
	}
	writing = 1;
	
	block = m_Streamer->Get(count);
	if (m_Streamer->m_SendBuffer->Available()>100*TS_SIZE) {		
		Dprintf("count:%d, avail: %d\n", count, m_Streamer->m_SendBuffer->Available());
		needlog = 1;
	}
	offset = 0;
	/*cStreamdevLiveStreamer *mS = (cStreamdevLiveStreamer*)m_Streamer;
	if ((count % TS_SIZE) != 0) 
	Dprintf("GOT %d %d bytes\n", mS->GetChannel()->Number(), count);*/

	if (block != NULL) {

		while (count>=TS_SIZE) {
			int written;
			int pkgsize = count;
			
			if (sel->Select(15000) == -1) {
				esyslog("ERROR: streamdev-server: couldn't send data: %m");
				break;
			}
			if (sel->CanWrite(*m_Socket)) {
				// SOCK_DGRAM indicates multicast
				if (m_Socket->Type() == SOCK_DGRAM) {
					// don't fragment multicast packets
					// max. payload on standard local ethernet is 1416 to 1456 bytes
					// and some STBs expect complete TS packets
					// so let's always limit to 7 * TS_SIZE = 1316
					if (pkgsize > 7 * TS_SIZE)
						pkgsize = 7 * TS_SIZE;
					else
						pkgsize -= pkgsize % TS_SIZE;
				}
				if ((written =     send (*m_Socket, block + offset, pkgsize, MSG_DONTWAIT )) == -1) {
					esyslog("ERROR: streamdev-server: couldn't send %d bytes: %m", pkgsize);
					writing = 0;
					return;
				}
				// statistics
				if (count > max)
					max = count;
				
				offset += written;
				count -= written;
				
				// less than one TS packet left:
				// delete what we've written so far and get next chunk
				//if (count < TS_SIZE) {
			}
		}
		m_Streamer->Del(offset);
		block = NULL;
	}
//	if (needlog)
	gettimeofday(&tim, NULL);
	t2 = (tim.tv_sec-1276142000)*1000000.0+tim.tv_usec;
	//Dprintf("%.3lf ms\n", t2-t1);
	
	//Dprintf("%c after: %.3lf ms, count:%d, deleted:%d, avail: %d\n", m_Streamer->m_SendBuffer->letter, t2-t1, count, offset, m_Streamer->m_SendBuffer->Available());
	t1=t2;

	writing = 0;
}


// --- cStreamdevStreamer -----------------------------------------------------

cStreamdevStreamer::cStreamdevStreamer(const char *Name):
		cThread(Name),
		m_Writer(NULL),
		m_RingBuffer(new cStreamdevBuffer(STREAMERBUFSIZE, TS_SIZE * 2,
		             true, "streamdev-streamer")),
		m_SendBuffer(new cStreamdevBuffer(WRITERBUFSIZE, TS_SIZE * 2))
{
// vld-timeout
	m_RingBuffer->SetTimeouts(0, 2);
	m_SendBuffer->SetTimeouts(2, 2);
	writing = 0;
	pcrPid = 0;
}

cStreamdevStreamer::~cStreamdevStreamer() 
{
	Dprintf("Desctructing streamer\n");
	delete m_RingBuffer;
	delete m_SendBuffer;
}

void cStreamdevStreamer::Start(cTBSocket *Socket) 
{
	Dprintf("start streamer\n");
	m_Socket = Socket;
	m_Writer = new cStreamdevWriter(Socket, this);
	Attach();
}

void cStreamdevStreamer::Activate(bool On) 
{
	if (On && !Active()) {
		Dprintf("activate streamer\n");
		m_Writer->Start();
		cThread::Start();
	}
}

void cStreamdevStreamer::Stop(void) 
{
	if (Running()) {
		Dprintf("stopping streamer\n");
		Cancel(3);
	}
	if (m_Writer) {
		Detach();
		DELETENULL(m_Writer);
	}
}

void cStreamdevStreamer::Restart(void)
{
	cStreamdevLiveStreamer *mS = (cStreamdevLiveStreamer*)this;
	esyslog("GOT restarting channel %d\n", mS->GetChannel()->Number());
	if (m_Writer) {
		Detach();
		DELETENULL(m_Writer);
	}
	sleep(2);
	m_Writer = new cStreamdevWriter(m_Socket, this);
	Attach();
}

void cStreamdevStreamer::Action(void) 
{
	while (Running()) {
		sleep(1);
/*		int got;
		uchar *block = m_RingBuffer->Get(got);

		if (block) {
			int count = Put(block, got);
			if (count)
				m_RingBuffer->Del(count);
		}*/
	}
}

void cStreamdevStreamer::ActionCall(void) 
{
	int got;

	if (writing) {
		Dprintf("streamer already writing\n");
		return;
	}

	writing = 1;
	uchar *block = m_RingBuffer->Get(got);
	
	if (block) {
		int count = Put(block, got);
		if (count)
			m_RingBuffer->Del(count);
	}
	writing = 0;
}

