/*
 *  $Id: streamer.h,v 1.11 2009/06/19 06:32:45 schmirl Exp $
 */
 
#ifndef VDR_STREAMDEV_STREAMER_H
#define VDR_STREAMDEV_STREAMER_H

#include <vdr/thread.h>
#include <vdr/ringbuffer.h>
#include <vdr/tools.h>
#include "tools/select.h"
#include <sys/time.h>
#include "common.h"

class cTBSocket;
class cStreamdevStreamer;

#ifndef TS_SIZE
#define TS_SIZE 188
#endif

#define STREAMERBUFSIZE (50000 * TS_SIZE)
#define WRITERBUFSIZE (30000 * TS_SIZE)

// --- cStreamdevBuffer -------------------------------------------------------

class cStreamdevBuffer: public cRingBufferLinear {
public:
	int letter;
	// make public
	void WaitForPut(void) { cRingBuffer::WaitForPut(); }
	// Always write complete TS packets
	// (assumes Count is a multiple of TS_SIZE)
	int PutTS(const uchar *Data, int Count);
	cStreamdevBuffer(int Size, int Margin = 0, bool Statistics = false, const char *Description = NULL);
};

inline int cStreamdevBuffer::PutTS(const uchar *Data, int Count)
{
	// Dprintf("%c", letter);
	int free = Free();
	if (free < Count)
		Count = free;

	Count -= Count % TS_SIZE;
	if (Count)
		Count = Put(Data, Count);
	else
		WaitForPut();
	return Count;
}

// --- cStreamdevWriter -------------------------------------------------------

class cStreamdevWriter: public cThread {
private:
	cStreamdevStreamer *m_Streamer;
	cTBSocket          *m_Socket;
	int writing;
	cTBSelect *sel;
	double t1;
	double t2;
	int fdw;
protected:
	virtual void Action(void);

public:
	cStreamdevWriter(cTBSocket *Socket, cStreamdevStreamer *Streamer);
	void ActionCall(void);
	virtual ~cStreamdevWriter();
};

// --- cStreamdevStreamer -----------------------------------------------------

class cStreamdevStreamer: public cThread {
private:
	cStreamdevWriter  *m_Writer;
	cStreamdevBuffer  *m_RingBuffer;
	cTBSocket *m_Socket;
	int writing;
	double t1;
	double t2;

protected:
	virtual void Action(void);

	bool IsRunning(void) const { return m_Writer; }

public:
	int pcrPid;
	cStreamdevBuffer  *m_SendBuffer;
	cStreamdevStreamer(const char *Name);
	virtual ~cStreamdevStreamer();

	virtual void Start(cTBSocket *Socket);
	virtual void Stop(void);
	void Restart(void);
	bool Abort(void);

	void Activate(bool On);
	int Receive(uchar *Data, int Length) { 
		
		//timeval tim;
		//Dprintf("XX%c", m_RingBuffer->letter); 
		int put = m_RingBuffer->PutTS(Data, Length);
		if (m_RingBuffer->Available() >= 7*TS_SIZE) {
			/*gettimeofday(&tim, NULL);
			t2 = (tim.tv_sec-1276142000)*1000000.0+tim.tv_usec;
			Dprintf("receive %.3lf ms\n", t2-t1);
			t1=t2;*/
			ActionCall();
		}
		return put;
	}
	void ReportOverflow(int Bytes) { m_RingBuffer->ReportOverflow(Bytes); }
	void ActionCall(void);
	
	virtual int Put(const uchar *Data, int Count) { 
		//Dprintf("LL%c", m_RingBuffer->letter); 
		int put = m_SendBuffer->PutTS(Data, Count); 
		//if (m_SendBuffer->Available() >= 7*TS_SIZE) {
		//	m_Writer->ActionCall();
		//}
		return put;
	}
	virtual uchar *Get(int &Count) { return m_SendBuffer->Get(Count); }
	virtual void Del(int Count) { m_SendBuffer->Del(Count); }

	virtual void Detach(void) {}
	virtual void Attach(void) {}
};

inline bool cStreamdevStreamer::Abort(void)
{
	return Active() && !m_Writer->Active();
}

#endif // VDR_STREAMDEV_STREAMER_H

